"""
This module provides view mappers for running views in asyncio.
"""
import asyncio

from pyramid.config.views import DefaultViewMapper
from pyramid.exceptions import ConfigurationError

from .helpers import synchronize, is_generator


class AsyncioMapperBase(DefaultViewMapper):
    """
    Base class for asyncio view mappers.
    """

    def run_in_coroutine_view(self, view):

        view = synchronize(view)

        def coroutine_view(context, request):

            # Trigger loading of post data to avoid synchronization problems
            # This must be done in a non-async context
            request.params.__getitem__ = request.params.__getitem__

            return view(context, request)

        return coroutine_view

    def run_in_executor_view(self, view):

        synchronizer = synchronize(strict=False)

        def executor_view(context, request):

            # Trigger loading of post data to avoid synchronization problems
            # This must be done in a non-async context
            request.params.__getitem__ = request.params.__getitem__

            try:
                # since we are running in a new thread,
                # remove the old wsgi.file_wrapper for uwsgi
                request.environ.pop('wsgi.file_wrapper')
            finally:
                exe = synchronizer(asyncio.get_event_loop().run_in_executor)
                return exe(None, view, context, request)

        return executor_view


class CoroutineMapper(AsyncioMapperBase):

    def __call__(self, view):
        original = view
        view = super().__call__(view)

        if is_generator(original) or is_generator(
            getattr(original, '__call__', None)
        ):
            view = asyncio.coroutine(view)
        elif not asyncio.iscoroutinefunction(original):
            raise ConfigurationError(
                'Non-coroutine {} mapped to coroutine.'.format(original)
            )

        return self.run_in_coroutine_view(view)


class ExecutorMapper(AsyncioMapperBase):

    def __call__(self, view):
        if asyncio.iscoroutinefunction(view) or asyncio.iscoroutinefunction(
            getattr(view, '__call__', None)
        ):
            raise ConfigurationError(
                'Coroutine {} mapped to executor.'.format(view)
            )
        view = super().__call__(view)
        return self.run_in_executor_view(view)


class CoroutineOrExecutorMapper(AsyncioMapperBase):

    def __call__(self, view):
        original = view
        while asyncio.iscoroutinefunction(view):
            try:
                view = view.__wrapped__  # unwrap coroutine
            except AttributeError:
                break

        view = super().__call__(view)

        if (
            asyncio.iscoroutinefunction(original) or
            is_generator(original) or
            is_generator(
                getattr(original, '__call__', None)
            )
        ):
            view = asyncio.coroutine(view)
            return self.run_in_coroutine_view(view)
        else:
            return self.run_in_executor_view(view)



import asyncio
import unittest

import greenlet
from pyramid.exceptions import ConfigurationError


class TestIsGenerator(unittest.TestCase):

    def test_regular_yield(self):
        from aiopyramid.helpers import is_generator

        def _sample():
            yield 1
            yield 2

        self.assertTrue(is_generator(_sample))

    def test_yield_from(self):
        from aiopyramid.helpers import is_generator

        def _placeholder():
            yield 6
            return 7

        def _sample():
            yield from _placeholder

        self.assertTrue(is_generator(_sample))

    def test_coroutine(self):
        from aiopyramid.helpers import is_generator

        @asyncio.coroutine
        def _sample():
            return 5

        self.assertTrue(is_generator(_sample))

    def test_false(self):
        from aiopyramid.helpers import is_generator

        def _sample():
            return "plain old function"

        self.assertFalse(is_generator(_sample))


class TestSpawnGreenlet(unittest.TestCase):

    def test_return_direct_result(self):
        from aiopyramid.helpers import spawn_greenlet

        def _return_4():
            return 4

        out = asyncio.get_event_loop().run_until_complete(
            spawn_greenlet(_return_4),
        )
        self.assertEqual(out, 4)

    def test_switch_direct_result(self):
        from aiopyramid.helpers import spawn_greenlet

        def _switch_4_return_5():
            this = greenlet.getcurrent()
            this.parent.switch(4)
            return 5

        out = asyncio.get_event_loop().run_until_complete(
            spawn_greenlet(_switch_4_return_5),
        )
        self.assertEqual(out, 4)

    def test_wait_on_future(self):
        from aiopyramid.helpers import spawn_greenlet

        future = asyncio.Future()

        def _switch_future():
            this = greenlet.getcurrent()
            this.parent.switch(future)
            return 5  # This should never get returned

        # the result is already set, but
        # spawn_greenlet will still need to yield from it
        future.set_result(4)
        out = asyncio.get_event_loop().run_until_complete(
            spawn_greenlet(_switch_future),
        )
        self.assertEqual(out, 4)


class TestRunInGreenlet(unittest.TestCase):

    def test_result(self):
        from aiopyramid.helpers import spawn_greenlet
        from aiopyramid.helpers import run_in_greenlet

        @asyncio.coroutine
        def _sample(pass_back):
            return pass_back

        def _greenlet():
            this = greenlet.getcurrent()
            future = asyncio.Future()
            message = 12
            sub_task = asyncio.async(
                run_in_greenlet(this, future, _sample, message),
            )
            this.parent.switch(sub_task)
            self.assertEqual(future.result(), message)
            return message + 1

        out = asyncio.get_event_loop().run_until_complete(
            spawn_greenlet(_greenlet),
        )
        self.assertEqual(13, out)

    def test_result_chain(self):
        from aiopyramid.helpers import spawn_greenlet
        from aiopyramid.helpers import run_in_greenlet

        @asyncio.coroutine
        def _sample(pass_back):
            return pass_back

        @asyncio.coroutine
        def _chain(pass_back):
            out = yield from _sample(pass_back)
            self.assertEqual(out, pass_back)
            return out - 1

        def _greenlet():
            this = greenlet.getcurrent()
            future = asyncio.Future()
            message = 12
            sub_task = asyncio.async(run_in_greenlet(
                this,
                future,
                _chain,
                message,
            ))
            this.parent.switch(sub_task)
            self.assertEqual(future.result(), message - 1)
            return message + 1

        out = asyncio.get_event_loop().run_until_complete(
            spawn_greenlet(_greenlet),
        )
        self.assertEqual(13, out)

    def test_exception(self):
        from aiopyramid.helpers import spawn_greenlet
        from aiopyramid.helpers import run_in_greenlet

        @asyncio.coroutine
        def _sample():
            raise KeyError

        def _greenlet():
            this = greenlet.getcurrent()
            future = asyncio.Future()
            sub_task = asyncio.async(
                run_in_greenlet(this, future, _sample),
            )
            this.parent.switch(sub_task)
            self.assertRaises(KeyError, future.result)

        asyncio.get_event_loop().run_until_complete(
            spawn_greenlet(_greenlet),
        )


class TestSynchronize(unittest.TestCase):

    @asyncio.coroutine
    def _sample(self, pass_back):
        return pass_back

    def _simple(self, pass_back):
        return pass_back

    def test_conversion(self):
        from aiopyramid.helpers import synchronize
        from aiopyramid.helpers import is_generator

        syncer = synchronize(strict=True)
        self.assertRaises(ConfigurationError, syncer, self._simple)
        self.assertFalse(is_generator(syncer(self._sample)))

    def test_scope_error(self):
        from aiopyramid.exceptions import ScopeError
        from aiopyramid.helpers import synchronize, spawn_greenlet

        synced = synchronize(self._sample)
        self.assertRaises(ScopeError, synced, 'val')

        five = asyncio.get_event_loop().run_until_complete(
            spawn_greenlet(synced, 5),
        )
        self.assertEqual(five, 5)

        synced = synchronize(self._sample, strict=False)
        self.assertTrue(asyncio.iscoroutine(synced('val')))

    def test_as_decorator(self):
        from aiopyramid.helpers import synchronize, spawn_greenlet
        from aiopyramid.exceptions import ScopeError

        @synchronize
        @asyncio.coroutine
        def _synced(pass_back):
            yield
            return pass_back

        self.assertRaises(ScopeError, _synced, 'val')
        twelve = asyncio.get_event_loop().run_until_complete(
            spawn_greenlet(_synced, 12),
        )
        self.assertEqual(twelve, 12)



import asyncio
import inspect
import functools
import logging

import greenlet
from pyramid.exceptions import ConfigurationError

from .exceptions import ScopeError

SCOPE_ERROR_MESSAGE = '''
Synchronized coroutine {} called in the parent
greenlet.

This is most likely because you called the synchronized
coroutine inside of another coroutine. You need to
yield from the coroutine directly without wrapping
it in aiopyramid.helpers.synchronize.

If you are calling this coroutine indirectly from
a regular function and therefore cannot yield from it,
then you need to run the first caller inside a new
greenlet using aiopyramid.helpers.spawn_greenlet.
'''

log = logging.getLogger(__name__)


def is_generator(func):
    """ Tests whether `func` is capable of becoming an `asyncio.coroutine`. """
    return (
        inspect.isgeneratorfunction(func) or
        isinstance(func, asyncio.Future) or
        inspect.isgenerator(func)
    )


@asyncio.coroutine
def spawn_greenlet(func, *args, **kwargs):
    """
    Spawns a new greenlet and waits on any `asyncio.Future` objects returned.

    This is used by the Gunicorn worker to proxy a greenlet within an `asyncio`
    event loop.
    """

    g = greenlet.greenlet(func)
    result = g.switch(*args, **kwargs)
    while True:
        if isinstance(result, asyncio.Future):
            result = yield from result
        else:
            break
    return result


@asyncio.coroutine
def run_in_greenlet(back, future, func, *args, **kwargs):
    """
    Wait for :term:`coroutine` func and switch back to the request greenlet
    setting any result in the future or an Exception where approrpiate.

    func is often a :term:`view callable`
    """
    try:
        result = yield from func(*args, **kwargs)
    except Exception as ex:
        future.set_exception(ex)
    else:
        future.set_result(result)
    finally:
        return back.switch()


def synchronize(*args, strict=True):
    """
    Decorator for transforming an async coroutine function into a regular
    function relying on the `aiopyramid` architecture to schedule
    the coroutine and obtain the result.

    .. code-block:: python

        @synchronize
        @asyncio.coroutine
        def my_coroutine():
            ... code that yields
    """

    def _wrapper(coroutine_func):
        if strict and not asyncio.iscoroutinefunction(coroutine_func):
            raise ConfigurationError(
                'Attempted to synchronize a non-coroutine {}.'.format(
                    coroutine_func
                )
            )

        @functools.wraps(coroutine_func)
        def _wrapped_coroutine(*args, **kwargs):

            this = greenlet.getcurrent()
            if this.parent is None:
                if strict:
                    raise ScopeError(
                        SCOPE_ERROR_MESSAGE.format(coroutine_func)
                    )
                else:
                    return coroutine_func(*args, **kwargs)
            else:
                future = asyncio.Future()
                sub_task = asyncio.async(
                    run_in_greenlet(
                        this,
                        future,
                        coroutine_func,
                        *args,
                        **kwargs
                    )
                )
                while not future.done():
                    this.parent.switch(sub_task)
                return future.result()

        return _wrapped_coroutine

    try:
        coroutine_func = args[0]
        return _wrapper(coroutine_func)
    except IndexError:
        return _wrapper


def spawn_greenlet_on_scope_error(func):
    """
    Wraps a callable handling any
    :class:`ScopeErrors <~aiopyramid.exceptions.ScopeError>` that may
    occur because the callable is called from inside of a :term:`coroutine`.

    If no :class:`~aiopyramid.exceptions.ScopeError` occurs, the callable is
    executed normally and return arguments are passed through, otherwise, when
    a :class:`~aiopyramid.exceptions.ScopeError` does occur, a coroutine to
    retrieve the result of the callable is returned instead.
    """

    @functools.wraps(func)
    def _run_or_return_future(*args, **kwargs):
        this = greenlet.getcurrent()
        # Check if we should see a ScopeError
        if this.parent is None:
            return spawn_greenlet(func, *args, **kwargs)
        else:
            try:
                return func(*args, **kwargs)
            except ScopeError:
                # ScopeError generated in multiple levels of indirection
                log.warn('Unexpected ScopeError encountered.')
                return spawn_greenlet(func, *args, **kwargs)

    return _run_or_return_future


def use_executor(*args, executor=None):
    """
    A decorator for running a callback in the executor.

    This is useful to provide a declarative style for converting some
    thread-based code to a :term:`coroutine`. It creates a :term:`coroutine`
    by running the wrapped code in a separate thread.

    """
    def _wrapper(callback):
        @functools.wraps(callback)
        @asyncio.coroutine
        def _wrapped_function(*args, **kwargs):
            loop = asyncio.get_event_loop()
            r = yield from loop.run_in_executor(
                executor,
                functools.partial(
                    callback,
                    *args,
                    **kwargs
                )
            )
            return r
        return _wrapped_function

    try:
        func = args[0]
        return _wrapper(func)
    except IndexError:
        return _wrapper



import asyncio


class WebsocketConnectionView:
    """ :term:`view callable` for websocket connections. """

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @asyncio.coroutine
    def __call__(self, ws):
        self.ws = ws
        yield from self.on_open()
        while True:
            message = yield from self.ws.recv()
            if message is None:
                yield from self.on_close()
                break
            yield from self.on_message(message)

    @asyncio.coroutine
    def send(self, message):
        yield from self.ws.send(message)

    @asyncio.coroutine
    def on_message(self, message):
        """
        Callback called when a message is received.
        Default is a noop.
        """
        pass

    @asyncio.coroutine
    def on_open(self):
        """
        Callback called when the connection is first established.
        Default is a noop.
        """

    @asyncio.coroutine
    def on_close(self):
        """
        Callback called when the connection is closed.
        Default is a noop.
        """
        pass



"""
Utilities for making :ref:`Pyramid <pyramid:index>` authentication
and authorization work with Aiopyramid.
"""


from .helpers import spawn_greenlet_on_scope_error, synchronize


def coroutine_callback_authentication_policy_factory(
    policy_class,
    coroutine=None,
    *args,
    **kwargs
):
    """
    Factory function for creating an AuthenticationPolicy instance that uses
    a :term:`coroutine` as a callback.

    :param policy_class: The AuthenticationPolicy to wrap.
    :param coroutine coroutine: If provided this is passed to
    the AuthenticationPolicy as the callback argument.

    Extra arguments and keyword arguments are passed to
    the AuthenticationPolicy, so if the AuthenticationPolicy expects
    a callback under another name, it is necessary to pass
    a :term:`synchronized coroutine` as an argument or keyword argument
    to this factory or use
    :class:`~aiopyramid.auth.CoroutineAuthenticationPolicyProxy` directly.

    This function is also aliased as
    :func:`aiopyramid.auth.authn_policy_factory`.
    """

    if coroutine:
        coroutine = synchronize(coroutine)
        policy = policy_class(callback=coroutine, *args, **kwargs)
    else:
        policy = policy_class(*args, **kwargs)
    return CoroutineAuthenticationPolicyProxy(policy)


authn_policy_factory = coroutine_callback_authentication_policy_factory


class CoroutineAuthenticationPolicyProxy:
    """
    This authentication policy proxies calls to another policy that uses
    a callback to retrieve principals. Because this callback may be a
    :term:`synchronized coroutine`, this class handles the case where the
    callback fails due to a :class:`~aiopyramid.exceptions.ScopeError` and
    generates the appropriate ``Aiopyramid`` architecture.
    """

    def __init__(self, policy):
        """
        :param class policy: The authentication policy to wrap.
        """

        self._policy = policy

    @spawn_greenlet_on_scope_error
    def remember(self, request, principal, **kwargs):
        return self._policy.remember(request, principal, **kwargs)

    @spawn_greenlet_on_scope_error
    def forget(self, request):
        return self._policy.forget(request)

    @spawn_greenlet_on_scope_error
    def unauthenticated_userid(self, request):
        return self._policy.unauthenticated_userid(request)

    @spawn_greenlet_on_scope_error
    def authenticated_userid(self, request):
        return self._policy.authenticated_userid(request)

    @spawn_greenlet_on_scope_error
    def effective_principals(self, request):
        return self._policy.effective_principals(request)



"""
Run pyramid app using asyncio
"""

from .config import CoroutineOrExecutorMapper


def includeme(config):
    """
    Setup the basic configuration to run :ref:`Pyramid <pyramid:index>`
    with :mod:`asyncio`.
    """

    config.set_view_mapper(CoroutineOrExecutorMapper)



"""
The aiopyramid.tweens module is deprecated. See example in the docs:
http://aiopyramid.readthedocs.org/en/latest/features.html#tweens.
"""

import asyncio
import warnings

from .helpers import synchronize

warnings.warn(__doc__, DeprecationWarning)


def coroutine_logger_tween_factory(handler, registry):
    """
    Example of an asynchronous tween that delegates a synchronous function to
    a child thread. This tween asynchronously logs all requests and responses.
    """

    # We use the synchronize decorator because we will call this
    # coroutine from a normal python context
    @synchronize
    # this is a coroutine
    @asyncio.coroutine
    def _async_print(content):
        # print doesn't really need to be run in a separate thread
        # but it works for demonstration purposes

        yield from asyncio.get_event_loop().run_in_executor(
            None,
            print,
            content
        )

    def coroutine_logger_tween(request):
        # The following calls are guaranteed to happen in order but they do not
        # block the event loop

        # print the request on the aio event loop without needing to say yield
        # at this point, other coroutines and requests can be handled
        _async_print(request)

        # get response, this should be done in this greenlet
        # and not as a coroutine because this will call
        # the next tween and subsequently yield if necessary
        response = handler(request)

        # print the response on the aio event loop
        _async_print(request)

        # return response after logging is done
        return response

    return coroutine_logger_tween



"""
The aiopyramid.traversal module is deprecated, use aiopyramid.helpers.synchronize instead.
See http://aiopyramid.readthedocs.org/en/latest/features.html#traversal.
"""  # NOQA

import asyncio
import warnings

from pyramid.traversal import (
    ResourceTreeTraverser as TraverserBase,
    is_nonstr_iter,
    split_path_info,
)
from pyramid.exceptions import URLDecodeError
from pyramid.interfaces import VH_ROOT_KEY
from pyramid.compat import decode_path_info

from .helpers import synchronize

SLASH = "/"

warnings.warn(__doc__, DeprecationWarning)


@synchronize
@asyncio.coroutine
def traverse(
    i,
    ob,
    view_selector,
    vpath_tuple,
    vroot_idx,
    vroot,
    vroot_tuple,
    root,
    subpath,
):
    """
    A version of :func:`pyramid.traversal.traverse` that expects `__getitem__`
    to be a :term:`coroutine`.
    """

    for segment in vpath_tuple:
        if segment[:2] == view_selector:
            return {
                'context': ob,
                'view_name': segment[2:],
                'subpath': vpath_tuple[i + 1:],
                'traversed': vpath_tuple[:vroot_idx + i + 1],
                'virtual_root': vroot,
                'virtual_root_path': vroot_tuple,
                'root': root,
            }
        try:
            getitem = ob.__getitem__
        except AttributeError:
            return {
                'context': ob,
                'view_name': segment,
                'subpath': vpath_tuple[i + 1:],
                'traversed': vpath_tuple[:vroot_idx + i + 1],
                'virtual_root': vroot,
                'virtual_root_path': vroot_tuple,
                'root': root,
            }

        try:
            tsugi = yield from getitem(segment)
        except KeyError:
            return {
                'context': ob,
                'view_name': segment,
                'subpath': vpath_tuple[i + 1:],
                'traversed': vpath_tuple[:vroot_idx + i + 1],
                'virtual_root': vroot,
                'virtual_root_path': vroot_tuple,
                'root': root,
            }
        if i == vroot_idx:
            vroot = tsugi
        ob = tsugi
        i += 1

    return {
        'context': ob,
        'view_name': "",
        'subpath': subpath,
        'traversed': vpath_tuple,
        'virtual_root': vroot,
        'virtual_root_path': vroot_tuple,
        'root': root
    }


class AsyncioTraverser(TraverserBase):
    """
    Traversal algorithm patched from the default traverser to execute
    __getitem__ as a coroutine.
    """

    def __call__(self, request):
        environ = request.environ
        matchdict = request.matchdict

        if matchdict is not None:

            path = matchdict.get('traverse', SLASH) or SLASH
            if is_nonstr_iter(path):
                # this is a *traverse stararg (not a {traverse})
                # routing has already decoded these elements, so we just
                # need to join them
                path = '/' + SLASH.join(path) or SLASH

            subpath = matchdict.get('subpath', ())
            if not is_nonstr_iter(subpath):
                # this is not a *subpath stararg (just a {subpath})
                # routing has already decoded this string, so we just need
                # to split it
                subpath = split_path_info(subpath)

        else:
            # this request did not match a route
            subpath = ()
            try:
                # empty if mounted under a path in mod_wsgi, for example
                path = request.path_info or SLASH
            except KeyError:
                # if environ['PATH_INFO'] is just not there
                path = SLASH
            except UnicodeDecodeError as e:
                raise URLDecodeError(e.encoding, e.object, e.start, e.end,
                                     e.reason)

        if VH_ROOT_KEY in environ:
            # HTTP_X_VHM_ROOT
            vroot_path = decode_path_info(environ[VH_ROOT_KEY])
            vroot_tuple = split_path_info(vroot_path)
            vpath = vroot_path + path
            vroot_idx = len(vroot_tuple) - 1
        else:
            vroot_tuple = ()
            vpath = path
            vroot_idx = - 1

        root = self.root
        ob = vroot = root

        if vpath == SLASH:
            vpath_tuple = ()
        else:
            i = 0
            view_selector = self.VIEW_SELECTOR
            vpath_tuple = split_path_info(vpath)
            return traverse(
                i,
                ob,
                view_selector,
                vpath_tuple,
                vroot_idx,
                vroot,
                vroot_tuple,
                root,
                subpath,
            )

        return {
            'context': ob,
            'view_name': "",
            'subpath': subpath,
            'traversed': vpath_tuple,
            'virtual_root': vroot,
            'virtual_root_path': vroot_tuple,
            'root': root
        }



import unittest
import asyncio

import greenlet

from aiopyramid.helpers import spawn_greenlet, run_in_greenlet


class TestTweens(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.get_event_loop()

    def _make_tweens(self):
        from pyramid.config.tweens import Tweens
        return Tweens()

    def _async_tween_factory(self, handler, registry):

        @asyncio.coroutine
        def _async_action():
            yield from asyncio.sleep(0.2)
            return 12

        def async_tween(request):
            this = greenlet.getcurrent()
            future = asyncio.Future()
            sub_task = asyncio.async(
                run_in_greenlet(this, future, _async_action),
            )
            self.assertIsInstance(sub_task, asyncio.Future)
            this.parent.switch(sub_task)
            self.assertEqual(future.result(), 12)
            return future

        return async_tween

    def _dummy_tween_factory(self, handler, registry):
        return handler

    def test_async_tween(self):
        out = self.loop.run_until_complete(
            spawn_greenlet(self._async_tween_factory(None, None), None),
        )
        self.assertEqual(out, 12)

    def test_example_tween(self):
        from aiopyramid.tweens import coroutine_logger_tween_factory
        out = self.loop.run_until_complete(
            spawn_greenlet(
                coroutine_logger_tween_factory(
                    lambda x: x,
                    None,
                ),
                None,
            )
        )
        self.assertEqual(None, out)

    def test_sync_tween_above(self):
        tweens = self._make_tweens()
        tweens.add_implicit('async', self._async_tween_factory)
        tweens.add_implicit('sync', self._dummy_tween_factory)
        chain = tweens(None, None)
        out = self.loop.run_until_complete(spawn_greenlet(chain, None))
        self.assertEqual(out, 12)

    def test_sync_tween_below(self):
        tweens = self._make_tweens()
        tweens.add_implicit('sync', self._dummy_tween_factory)
        tweens.add_implicit('async', self._async_tween_factory)
        chain = tweens(None, None)
        out = self.loop.run_until_complete(spawn_greenlet(chain, None))
        self.assertEqual(out, 12)

    def test_sync_both(self):
        tweens = self._make_tweens()
        tweens.add_implicit('sync', self._dummy_tween_factory)
        tweens.add_implicit('async', self._async_tween_factory)
        tweens.add_implicit('sync', self._dummy_tween_factory)
        chain = tweens(None, None)
        out = self.loop.run_until_complete(spawn_greenlet(chain, None))
        self.assertEqual(out, 12)


class TestTweensGunicorn(unittest.TestCase):

    """ Test aiopyramid tweens gunicorn style. """
    # TODO write tests that rely on subtasks



import asyncio

import pytest
from pyramid import testing
from aiopyramid.helpers import spawn_greenlet, synchronize


@pytest.yield_fixture
def web_request():
    request = testing.DummyRequest()
    yield request


class TestAuthentication:

    @pytest.yield_fixture
    def wrapped_policy(self):
        from pyramid.authentication import CallbackAuthenticationPolicy
        from aiopyramid.auth import authn_policy_factory

        @asyncio.coroutine
        def callback(userid, request):
            yield from asyncio.sleep(0.1)
            return ['test_user']

        class TestAuthenticationPolicy(CallbackAuthenticationPolicy):
            def __init__(self, callback):
                self.callback = callback
                self.debug = True

            def unauthenticated_userid(self, request):
                return 'theone'

        yield authn_policy_factory(TestAuthenticationPolicy, callback)

    def call_authn_policy_methods(self, policy, request):
        assert policy.unauthenticated_userid(request) == 'theone'
        assert policy.authenticated_userid(request) == 'theone'
        assert policy.effective_principals(request) == [
            'system.Everyone',
            'system.Authenticated',
            'theone',
            'test_user',
        ]

    @asyncio.coroutine
    def yield_from_authn_policy_methods(self, policy, request):
        assert (yield from policy.unauthenticated_userid(request)) == 'theone'
        assert (yield from policy.authenticated_userid(request)) == 'theone'
        assert (yield from policy.effective_principals(request)) == [
            'system.Everyone',
            'system.Authenticated',
            'theone',
            'test_user',
        ]

    def test_wrapper_in_sync(self, wrapped_policy, web_request):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(spawn_greenlet(
            self.call_authn_policy_methods,
            wrapped_policy,
            web_request,
        ))

    def test_wrapper_in_coroutine(self, wrapped_policy, web_request):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(spawn_greenlet(
            synchronize(self.yield_from_authn_policy_methods),
            wrapped_policy,
            web_request,
        ))



import asyncio
import inspect
import functools

import websockets
import gunicorn  # noqa

from pyramid.response import Response

from aiopyramid.config import AsyncioMapperBase


def _connection_closed_to_none(func):
    """
    A backwards compatibility shim for websockets 3+. We need to
    still return `None` rather than throwing an exception in order
    to unite the interface with uWSGI even though the exception is
    more Pythonic.
    """

    @asyncio.coroutine
    @functools.wraps(func)
    def _connection_closed_to_none_inner(*args, **kwargs):
        try:
            msg = yield from func(*args, **kwargs)
        except websockets.exceptions.ConnectionClosed:
            msg = None

        return msg

    return _connection_closed_to_none_inner


def _use_bytes(func):
    """
    Encodes strings received from websockets to bytes to
    provide consistency with uwsgi since we don't have access
    to the raw WebsocketFrame.
    """

    @asyncio.coroutine
    @functools.wraps(func)
    def _use_bytes_inner(*args, **kwargs):
        data = yield from func(*args, **kwargs)
        if isinstance(data, str):
            return str.encode(data)
        else:
            return data

    return _use_bytes_inner


class HandshakeInterator:

    def __init__(self, app_iter):
        self.content = list(app_iter)
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.content[self.index]
        except IndexError:
            raise StopIteration
        finally:
            self.index += 1


class SwitchProtocolsResponse(Response):
    """Upgrade from a WSGI connection with the WebSocket handshake."""

    def __init__(self, environ, switch_protocols):
        super().__init__()
        self.status_int = 101

        http_1_1 = environ['SERVER_PROTOCOL'] == 'HTTP/1.1'

        def get_header(k):
            return environ['HTTP_' + k.upper().replace('-', '_')]

        key = websockets.handshake.check_request(get_header)

        if not http_1_1 or key is None:
            self.status_int = 400
            self.content = "Invalid WebSocket handshake.\n"
        else:
            set_header = self.headers.__setitem__
            websockets.handshake.build_response(set_header, key)
            self.app_iter = HandshakeInterator(self.app_iter)
            self.app_iter.close = switch_protocols


class WebsocketMapper(AsyncioMapperBase):

    use_bytes = False

    def launch_websocket_view(self, view):

        def websocket_view(context, request):

            if inspect.isclass(view):
                view_callable = view(context, request)
            else:
                view_callable = view

            @asyncio.coroutine
            def _ensure_ws_close(ws):
                if WebsocketMapper.use_bytes:
                    ws.recv = _use_bytes(ws.recv)

                ws.recv = _connection_closed_to_none(ws.recv)

                yield from view_callable(ws)
                yield from ws.close()

            def switch_protocols():
                # TODO: Determine if there is a more standard way to do this
                ws_protocol = websockets.WebSocketCommonProtocol()
                transport = request.environ['async.writer'].transport

                http_protocol = request.environ['async.protocol']
                http_protocol.connection_lost(None)

                transport._protocol = ws_protocol
                ws_protocol.connection_made(transport)
                asyncio.async(_ensure_ws_close(ws_protocol))

            response = SwitchProtocolsResponse(
                request.environ,
                switch_protocols,
            )
            # convert iterator to avoid eof issues
            response.body = response.body

            return response

        return websocket_view

    def __call__(self, view):
        """ Accepts a view_callable class. """
        return self.launch_websocket_view(view)




from .exceptions import WebsocketClosed


def ignore_websocket_closed(app):
    """ Wrapper for ignoring closed websockets. """

    def _call_app_ignoring_ws_closed(environ, start_response):
        try:
            return app(environ, start_response)
        except WebsocketClosed as e:
            if e.__cause__:
                raise
            return ('')
    return _call_app_ignoring_ws_closed



import unittest
import asyncio

from pyramid.traversal import traverse

from aiopyramid.helpers import spawn_greenlet, synchronize


class DummyResource:
    """ Dummy resource for testing async traversal. """

    def __init__(self, name, parent):
        self.__name__ = name
        self.__parent__ = parent
        self._dict = {}

    @synchronize
    @asyncio.coroutine
    def __getitem__(self, key):
        yield from asyncio.sleep(0.1)
        return self._dict[key]

    def __setitem__(self, key, value):
        self._dict[key] = value

    def add_child(self, name, klass):
        resource = klass(name=name, parent=self)
        self[name] = resource


class TestTraversal(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.get_event_loop()

    def test_async_traversed_length(self):
        resource = DummyResource('root', None)
        resource.add_child('cat', DummyResource)
        out = self.loop.run_until_complete(
            spawn_greenlet(traverse, resource, ['cat']),
        )
        self.assertEqual(len(out['traversed']), 1)

    def test_async_root(self):
        resource = DummyResource('root', None)
        resource.add_child('cat', DummyResource)
        out = self.loop.run_until_complete(
            spawn_greenlet(traverse, resource, ['']),
        )
        self.assertTrue(out.get('root') == out.get('context'))

    def test_async_depth(self):
        resource = DummyResource('root', None)
        resource.add_child('cat', DummyResource)
        out = self.loop.run_until_complete(
            spawn_greenlet(traverse, resource, ['cat']),
        )
        out['context'].add_child('dog', DummyResource)
        out = self.loop.run_until_complete(
            spawn_greenlet(traverse, resource, ['cat', 'dog']),
        )
        self.assertListEqual(list(out['traversed']), ['cat', 'dog'])

    def test_async_view_name(self):
        resource = DummyResource('root', None)
        resource.add_child('cat', DummyResource)
        out = self.loop.run_until_complete(
            spawn_greenlet(traverse, resource, ['cat', 'mouse']),
        )
        self.assertListEqual(list(out['traversed']), ['cat'])
        self.assertEqual(out['view_name'], 'mouse')



import inspect
import asyncio
from contextlib import suppress

import greenlet

from aiopyramid.config import AsyncioMapperBase
from aiopyramid.helpers import run_in_greenlet
from aiopyramid.websocket.exceptions import WebsocketClosed

try:
    import uwsgi
except ImportError:
    pass


def uwsgi_recv_msg(g):
    g.has_message = True
    g.switch()


class UWSGIWebsocket:

    def __init__(self, back, q_in, q_out):
        self.back = back
        self.q_in = q_in
        self.q_out = q_out
        self.open = True

    @asyncio.coroutine
    def recv(self):
        return (yield from self.q_in.get())

    @asyncio.coroutine
    def send(self, message):
        yield from self.q_out.put(message)
        self.back.switch()

    @asyncio.coroutine
    def close(self):
        yield from self.q_in.put(None)
        self.back.throw(WebsocketClosed)


class UWSGIWebsocketMapper(AsyncioMapperBase):

    use_str = True

    def launch_websocket_view(self, view):

        def websocket_view(context, request):
            uwsgi.websocket_handshake()
            this = greenlet.getcurrent()
            this.has_message = False
            q_in = asyncio.Queue()
            q_out = asyncio.Queue()

            # make socket proxy
            if inspect.isclass(view):
                view_callable = view(context, request)
            else:
                view_callable = view
            ws = UWSGIWebsocket(this, q_in, q_out)

            # start monitoring websocket events
            asyncio.get_event_loop().add_reader(
                uwsgi.connection_fd(),
                uwsgi_recv_msg,
                this
            )

            # NOTE: don't use synchronize because we aren't waiting
            # for this future, instead we are using the reader to return
            # to the child greenlet.

            future = asyncio.Future()
            asyncio.async(
                run_in_greenlet(this, future, view_callable, ws)
            )

            # switch to open
            this.parent.switch()

            while True:
                if future.done():
                    if future.exception() is not None:
                        raise WebsocketClosed from future.exception()
                    raise WebsocketClosed

                # message in
                if this.has_message:
                    this.has_message = False
                    try:
                        msg = uwsgi.websocket_recv_nb()
                    except OSError:
                        msg = None

                    if UWSGIWebsocketMapper.use_str:
                        with suppress(Exception):
                            print('howdy')
                            msg = bytes.decode(msg)

                    if msg or msg is None:
                        q_in.put_nowait(msg)

                # message out
                if not q_out.empty():
                    msg = q_out.get_nowait()
                    try:
                        uwsgi.websocket_send(msg)
                    except OSError:
                        q_in.put_nowait(None)

                this.parent.switch()

        return websocket_view

    def __call__(self, view):
        """ Accepts a view_callable class. """
        return self.launch_websocket_view(view)

UWSGIWebsocketMapper.use_str = False



from pyramid.scaffolds import PyramidTemplate


class AioStarterTemplate(PyramidTemplate):
    _template_dir = 'aio_starter'
    summary = 'Pyramid project using asyncio'


class AioWebsocketTemplate(PyramidTemplate):
    _template_dir = 'aio_websocket'
    summary = 'Aiopyramid project with websocket-based view'



import os
import sys
from setuptools import setup, find_packages

py_version = sys.version_info[:2]
if py_version < (3, 3):
    raise Exception("aiopyramid requires Python >= 3.3.")

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.rst')) as readme:
    README = readme.read()
with open(os.path.join(here, 'CHANGES.rst')) as changes:
    CHANGES = changes.read()


requires = [
    'pyramid',
    'greenlet',
]

if py_version < (3, 4):
    requires.append('asyncio')

setup(
    name='aiopyramid',
    version='0.3.4',
    description='Tools for running pyramid using asyncio.',
    long_description=README + '\n\n\n\n' + CHANGES,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Intended Audience :: Developers",
        "License :: Repoze Public License",
    ],
    author='Jason Housley',
    author_email='housleyjk@gmail.com',
    url='https://github.com/housleyjk/aiopyramid',
    keywords='pyramid asyncio greenlet wsgi',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    extras_require={
        'gunicorn': ['gunicorn>=19.1.1', 'aiohttp', 'websockets'],
    },
    license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
    entry_points="""\
    [pyramid.scaffold]
    aio_starter=aiopyramid.scaffolds:AioStarterTemplate
    aio_websocket=aiopyramid.scaffolds:AioWebsocketTemplate
    """
)



__all__ = []

try:
    from .uwsgi import *
    __all__.append('UWSGIWebsocketMapper')
except ImportError:
    pass

try:
    from .gunicorn import *
    __all__.append('WebsocketMapper')
except ImportError:
    pass



import asyncio
import time
import io

from gunicorn.workers.gaiohttp import AiohttpWorker
from aiohttp.wsgi import WSGIServerHttpProtocol

from aiopyramid.helpers import (
    spawn_greenlet,
    spawn_greenlet_on_scope_error,
    synchronize,
)


class AiopyramidHttpServerProtocol(WSGIServerHttpProtocol):

    @asyncio.coroutine
    def handle_request(self, message, payload):
        """ Patched from aiohttp. """
        now = time.time()

        if self.readpayload:
            wsgiinput = io.BytesIO()
            wsgiinput.write((yield from payload.read()))
            wsgiinput.seek(0)
            payload = wsgiinput
        else:
            # allow read to be called from a synchronous context
            payload.read = synchronize(payload.read)
            payload.read = spawn_greenlet_on_scope_error(payload.read)

        environ = self.create_wsgi_environ(message, payload)
        # add a reference to this for switching protocols
        environ['async.protocol'] = self

        response = self.create_wsgi_response(message)

        riter = yield from spawn_greenlet(
            self.wsgi,
            environ,
            response.start_response
        )

        resp = response.response
        try:
            for item in riter:
                if isinstance(item, asyncio.Future):
                    item = yield from item
                yield from resp.write(item)

            yield from resp.write_eof()
        finally:
            if hasattr(riter, 'close'):
                riter.close()

        if resp.keep_alive():
            self.keep_alive(True)

        self.log_access(
            message, environ, response.response, time.time() - now)


class AsyncGunicornWorker(AiohttpWorker):

    def factory(self, wsgi, *args):
        proto = AiopyramidHttpServerProtocol(
            wsgi,
            loop=self.loop,
            readpayload=True,
            log=self.log,
            keep_alive=self.cfg.keepalive,
            access_log=self.log.access_log,
            access_log_format=self.cfg.access_log_format)
        return self.wrap_protocol(proto)
